# backend.py
#
# (c) Copyright 2010 Michael Towers (larch42 at googlemail dot com)
#
# This file is part of the larch project.
#
#    larch is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    larch is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with larch; if not, write to the Free Software Foundation, Inc.,
#    51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
#----------------------------------------------------------------------------
# 2010.11.06

# Mount point for the installation root partition
IBASE = '/tmp/larchin_install'
# File to prevent two instances of larchin from running
LOCKFILE = '/tmp/larchin_lock'
# Log file for installation process
LOGFILE = '/tmp/larchin_log'


import sys, traceback, os, re, __builtin__
__builtin__.base_dir = os.path.dirname(os.path.dirname(
        os.path.realpath(__file__)))

from subprocess import Popen, PIPE, STDOUT
import json as serialize

sys.path.append(os.path.dirname(base_dir))
from liblarch.translation import i18n_module, lang
__builtin__._ = i18n_module(base_dir, 'larch')
__builtin__.lang = lang
# Run subprocesses without i18n in case the output is parsed.
os.environ["LANGUAGE"] = "C"

import locale
# Try to work around encoding problems
encoding = locale.getdefaultlocale()[1]


def _out(text, force=False):
    """Send the string to standard output.
    How it is output depends on the '-s' command line option (whether the
    script is being run on the console or as a subprocess of another script).
    In the latter case the text will be slightly encoded - to avoid newline
    characters - and sent as a single unit.
    Otherwise output the lines as they are, but all lines except
    the first get a '--' prefix.
    """
    lines = text.encode('utf-8').splitlines()
    if _log and not text.startswith('>-'):
        # Don't log the progress report lines
        _log.write(lines[0] + '\n')
        for l in lines[1:]:
            _log.write('--' + l + '\n')

    if force or not _quiet:
        if _options.slave:
            sys.stdout.write(serialize.dumps(text) + '\n')
        else:
            prefix = ''
            for line in lines:
                sys.stdout.write(prefix + line + '\n')
                prefix = '--'
        sys.stdout.flush()


def comment(text):
    _out('##' + text)


def report(text, key='#>'):
    _out(key + text)


def query_yn(message, default=False):
    _out('?>' + message)
    result = default

    if _options.slave:
        result = (raw_input().strip() == '??YES')

    elif not _quiet:
        prompt = _("Yes:y|No:n")
        py, pn = prompt.split('|')
        respy = py.lower().split(':')
        respn = pn.lower().split(':')
        while True:
            resp = raw_input("   [ %s ]: " % prompt).strip().lower()
            if resp in respy:
                result = True
                break
            if resp in respn:
                result = False
                break

    _out('#>%s' % ('Yes' if result else 'No'))
    return result


def query(message, default=''):
    _out('?<%s>%s' % (default if _options.slave else '', message))
    result = default
    if _options.slave:
        res = raw_input().strip()
        if res != '??*':
            if not res.startswith('??'):
                errout(_("Illegal return value: %s") % res, 1)
            result = res[2:]

    elif (not _quiet) or (not default):
        while True:
            resp = raw_input('>>> %s' % ('[Default: %s ]: ' % default)
                    if default else '').strip()
            if resp:
                result = resp
                break
            elif default:
                break

    _out('#>%s' % result)
    return result


def run_capture(*cmd, **kargs):
    kargs['stdout'] = PIPE
#    kargs['stderr'] = PIPE # or STDOUT?
    p = Popen(cmd, **kargs)
    res = p.communicate()[0].decode(encoding)
    return (p.returncode == 0, res)
__builtin__.run_capture = run_capture


def init(options, app_quit=None):
    global quit_function, _log, _options, _quiet
    _options = options
    _quiet = False if _options.slave else _options.quiet
    _log = None
    quit_function = app_quit if app_quit else sys.exit

    def debug(text):
        sys.stderr.write("DEBUG: " + text.strip() + "\n")
        sys.stderr.flush()
    __builtin__.debug = debug

    def errout(message="ERROR", _quit=0):
        _out('!>' + message, True)
        if _quit:
            scripts.killall()
            scripts.wait()
            sys.exit(_quit)
    __builtin__.errout = errout

    # Catch all unhandled errors.
    def errortrap(type, value, tb):
        etext = "".join(traceback.format_exception(type, value, tb))
        for line in etext.splitlines():
            _out('!+' + line, True)
        errout(_("Something went wrong, quitting (see log)"), 100)
    sys.excepthook = errortrap

    # Check no other larchin instance is running
    if os.path.isfile(LOCKFILE):
        if not query_yn(_(
                "Another instance of the installer seems to be running already.\n"
                "If you are absolutely sure no other installer instance is\n"
                "running, you may continue. Otherwise you should cancel.\n"
                "\nShall I continue?")):
            quit_function(102)
    else:
        writefile("", LOCKFILE)
    _log = open(LOGFILE, 'w')

    __builtin__.scripts = Scripts(base_dir + "/scripts/")


def readdata(filename):
    return readfile(base_dir + '/data/' + filename)


def readfile(fpath):
    """Read the given file. Return the read text as unicode.
    If there is an error, report it and return None.
    """
    try:
        with open(fpath) as fh:
            return fh.read().decode(encoding)
    except:
        errout(_("Couldn't read file: %s\n  %s") % (fpath, traceback.format_exc()))
        return None
    return text
__builtin__.readfile = readfile


def writefile(text, path):
    try:
        pd = os.path.dirname(path)
        if not os.path.isdir(pd):
            os.makedirs(pd)
        fh = None
        fh = open(path, 'w')
        fh.write(text)
        return True
    except:
        return False
    finally:
        if fh:
            fh.close()


class Scripts:
    def __init__(self, scriptdir):
        self.scriptdir = scriptdir
        self.processes = []

    def start_script(self, *args):
        report(' '.join(args), '#+')
        p = Popen((self.scriptdir + args[0],) + args[1:],
                stdout=PIPE, stderr=STDOUT)
        self.processes.append(p)
        p.scriptname = args[0]
        p.output = None
        return p

    def end_script(self, p):
        p.wait()
        report(p.scriptname, '#-')
        self.processes.remove(p)
        return p.returncode==0

    def run(self, *args):
        p = self.start_script(*args)
        return self._runscript(p)

    def script(self, *args):
        p = self.start_script(*args)
        if self._runscript(p):
            return p.outputtext
        else:
            return None

    def _runscript(self, p):
        p.outputtext = ""
        while True:
            line = p.stdout.readline()
            if not line:
                break
            p.outputtext += line
            report(line.rstrip())
        # The process has ended.
        return self.end_script(p)

    def killall(self):
        for p in self.processes:
            Popen(["pkill", "-g", str(p.pid)], stdout=PIPE).communicate()

    def wait(self):
        for p in list(self.processes):
            p.wait()
