#!/usr/bin/env python2
# -*- coding: UTF-8 -*-
#
# partitions.py - partition handling
#
# (c) Copyright 2010 Michael Towers (larch42 at googlemail dot com)
#
# This file is part of the larch project.
#
#    larch is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    larch is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with larch; if not, write to the Free Software Foundation, Inc.,
#    51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
#----------------------------------------------------------------------------
# 2010.11.06

import sys, os, re, atexit

import backend
HOMESIZE_MIN = 1000     # MB
ROOTSCALE_MIN = 5       # min root size is this x the size of the live system
LIVEPLUS_MIN = 200      # MB, min extra live space to allocate
LIVEPLUS_DEF = 500      # MB, default extra live space to allocate
FORMATCOMMAND = ['mkfs.ext4', '-L'] # default formatting command
LIVESYS = '/.livesys/medium'    # mount point of live medium

def getmemsize():
    """Return the size of the installed memory in MB (10^6 B).
    """
    # The first line of /proc/meminfo is like 'MemTotal:        3090688 kB'
    for line in readfile('/proc/meminfo').splitlines():
        if line.startswith('MemTotal:'):
            return int(line.split()[1]) * 1024 / 10**6      # in MB
    return 0


def getfillsize(part_mp):
    """Get the usage in MB (10^6 B) for the given partition or mount-point.
    """
    for line in run_capture('df', '-B', 'MB')[1].splitlines():
        ls = line.split()
        if (ls[0] == part_mp) or (ls[-1] == part_mp):
            return mb2num(ls[2])
    return None


def mb2num(mb):
    """Convert string MB sizes from readpartitions (which uses parted) to a number.
    """
    return int(mb.replace(',', '.')[:-2])


def getpartsize(dev):
    """Return the size in MB of the given partition.
    """
    d = os.path.basename(dev)
    for line in readfile('/proc/partitions').splitlines():
        ls = line.split()
        if ls and (ls[-1] == d):
            return int(ls[2]) * 1024 / 10**6
    return None



class Partitions:
    def __init__(self, dev, mp):
        self.device = dev
        self.mountpoint = mp
        self.formatcommands = {}    # mapping from partition label to format command

# There is some duplication of effort here - maybe move the mounting stuff
# out of backend.
        self._mounts = []

        atexit.register(self.sys_quit)

        # Determine available free space (also provides partition info)
        res = self.freespace()
        if res:
            errout(res, 1)
        # Get size of installed memory (for size of swap partition)
        self.memsize = getmemsize()


    def setformatmap(self, fmap):
        self.formatcommands = fmap


    def autolive(self, label, size, device=None):
        """Automatically (create and) format a partition for installing
        the live system 'frugally' - as a sort of rescue partition.
        If device is passed or a partition with the given label exists
        already, this partition will be used, otherwise a new one will be
        allocated.
        """
        # Determine and check live size
        livepart = run_capture('blkid', '-L', label)[1].strip() if label else None
        if device:
            if livepart and (livepart != device):
                return _("Partition with label '%s' already exists") % label
            livepart = device
        livesys_size = getfillsize(LIVESYS)
        livesize_min = livesys_size + LIVEPLUS_MIN
        if livepart:
            ls = getpartsize(livepart)
            if not ls:
                return _("Couldn't determine size of partition '%s'") % livepart
            if (ls < size) or (ls < livesize_min):
                return _("Existing live partition (%s) is too small") % livepart
#TODO:
# Reallocate partition to ensure correct id?

        else:
            if size < 0:
                size = livesys_size + LIVEPLUS_DEF
            elif size < livesize_min:
                return _("Size requested for live system is too small")
            # Allocate new live partition
            res = self.newpart(size)
            if res:
                return res
            livepart = self.device + str(self.maxpart)

        res = self.format(label, livepart)
        if res:
            return res


    def autopart(self,
            swaplabel, swapsize,
            rootlabel, rootsize, rootsize_default,
            homelabel, homesize):
        """Create partitions for the swap, root and home filesystems.
        If swapsize or homesize is 0, the corresponding partition
        will not be created.
        The sizes given here are specified as options. If they are <0 that
        means no value was passed and the default should be used. The default
        size for the root partition is also passed in, 'swap' is the same size
        as the installed memory (by default) and 'home' is the remaining space
        after the other partitions have been allocated. If neither the root
        space nor the home space is explicitly specified, home will always
        be at least as large as root, down to a certain limit, where the
        allocation will fail.
        """
        free = self.free_mb

        swappart = run_capture('blkid', '-L', swaplabel)[1].strip() if swaplabel else None
        rootpart = run_capture('blkid', '-L', rootlabel)[1].strip()
        homepart = run_capture('blkid', '-L', homelabel)[1].strip() if homelabel else None

        rootsize_min = getfillsize(LIVESYS) * ROOTSCALE_MIN
        if rootsize_default < rootsize_min:
            rootsize_default = rootsize_min

        if rootpart:
            rs = getpartsize(rootpart)
            if (rs < rootsize) or (rs < rootsize_min):
                return _("Existing main partition (%s) is too small") % rootpart
#TODO:
# Reallocate partition to ensure correct id?
            self.format(rootlabel, rootpart)
            rootsize = 0

        if homepart:
            hs = getpartsize(homepart)
            if (hs < homesize) or (hs < HOMESIZE_MIN):
                return _("Existing user partition (%s) is too small") % homepart
#TODO:
# Reallocate partition to ensure correct id?
            self.format(homelabel, homepart)
            homesize = 0

        # Determine and check swap size
        if swappart:
            if getpartsize(swappart) < swapsize:
                return _("Existing swap partition (%s) is too small") % swappart
            if self.getfstype(swappart) != 'swap':
#TODO:
# Reallocate partition to ensure correct id?
                self.format(swaplabel, swappart, swap=True)

            swapsize = 0
        else:
            if swapsize < 0:
                swapsize = self.memsize * 2
            free -= swapsize

        # Determine root and home sizes
        if homesize < 0:
            if rootsize < 0:
                homesize = free - rootsize_default
                if homesize < rootsize_default:
                    rootsize = free / 2
                    homesize = free - rootsize
                else:
                    rootsize = rootsize_default
            else:
                homesize = free - rootsize
                if homesize < HOMESIZE_MIN:
                    return _("Too little space available for '/home' partition")

        else:
            free -= homesize
            if rootsize < 0:
                rootsize = rootsize_default if rootsize_default <= free else free
            elif rootsize > free:
                    return _("Too little space available for '/' partition")

        if rootsize and (rootsize < rootsize_min):
            return _("Too little space allocated for '/' partition")

        # Allocate and format new swap partition
        if swapsize:
            res = self.newpart(swapsize, swap=True)
            if res:
                return res
            res = self.format(swaplabel, self.device + str(self.maxpart), swap=True)
            if res:
                return res

        # Allocate and format new root partition
        if rootsize:
            res = self.newpart(rootsize)
            if res:
                return res
            res = self.format(rootlabel, self.device + str(self.maxpart))
            if res:
                return res

        # Allocate and format new home partition
        if homesize:
            res = self.newpart(homesize)
            if res:
                return res
            res = self.format(homelabel, self.device + str(self.maxpart))
            if res:
                return res


    def freespace(self):
        """Determine the current maximum partition number and the amount of
        free space (in MB) at the end of the device. The partitioning
        info for the device is reread.
        """
        fail = _("Couldn't determine free space on %s") % self.device
        res = self._readpartitions()
        if not self.devinfo:
            if res != 'NOPARTITIONTABLE':
                return res if res else fail
#TODO: Do I need a dialog here?
# Or just go ahead and create a partition table?
            ok, rdpart = run_capture('parted', self.device,
                    'mklabel', 'msdos')
            if not ok:
                return (_("Couldn't create partition table on %s:\n  ")
                        % self.device) + rdpart
                #T:Konnte Partitionstabelle nicht erstellen

            res = self._readpartitions()
            if not self.devinfo:
                return fail

        self.partlist = self.devinfo[3]
        self.maxpart = 0
        self.start_mb = 0
        for p in self.partlist:
            if p[0] == u'-':
                continue
            if p[0] > self.maxpart:
                self.maxpart = p[0]
                self.start_mb = mb2num(p[5])  # the end of the partition
            elif mb2num(p[5]) > self.start_mb:
                return _("Current partitioning on %s cannot be handled") % self.device
                #T:Aktuelle Partitionierung kann nicht bearbeitet werden
        self.free_mb = mb2num(self.devinfo[1]) - self.start_mb
        return None


    def _readpartitions(self):
        """The list of partition returned here, in self.devinfo, is ordered
        in the same way as parted output. If the partitions are not in
        order of increasing sector numbers this will probably cause problems.
        """
        self.devinfo = None
        dev = os.path.basename(self.device)
        # Get list of 'uninteresting' (essentially, 'extended') partitions.
        eparts = []
        for line in readfile('/proc/partitions').splitlines():
            m = re.search(ur'.* 1 (sd[a-z][0-9]+)', line)
            if m:
                eparts.append(m.group(1))

        # Now get list of all partitions, filtering out the 'uninteresting' ones
        partinfo = []
        ok, rdpart = run_capture('parted', '-m', self.device, 'unit', 'MB', 'print', 'free')
        if not ok:
            if (self.device + ': unrecognised disk label') in rdpart:
                return 'NOPARTITIONTABLE'
            else:
                return _("Couldn't get disk information:\n  ") + rdpart
                #T:Konnte Festplatteinformationen nicht lesen
        for line in rdpart.splitlines():
            line = line.rstrip(';')
            if line[0] in '123456789':
                info = line.split(':')
                partsize = info[3]
                partstart = info[1]
                partend = info[2]
                if info[-1] == 'free':
                    partnum = u'-'
                    partfs = _("Unused")
                    #T:Unbenutzt
                    flags = u''
                else:
                    partnum = info[0]
                    if (dev + partnum) in eparts:
                        continue
                    #partfs = info[4]   # rather use blkid?
                    partfs = run_capture('blkid', '-s', 'TYPE', '-o', 'value',
                            self.device+partnum)[1].strip()
                    flags = info[6]
                partinfo.append((partnum, partsize, partfs, flags, partstart, partend))

            if line.startswith('/dev/'):
                info = line.split(':')
                # (device (without '/dev/'), size, model, partition-list,
                #       disk-label (msdos/gpt), sector size (logical/physical))
                self.devinfo = (dev, info[1], '%s (%s)' % (info[6], info[2]),
                        partinfo, info[5], info[3], info[4])

        return None


    def removepartitions(self, partition0=1):
        """Remove all partitions starting at the given one.
        """
        fails = ''
        while self.maxpart >= partition0:
            ok, output = run_capture('parted', self.device, 'rm', str(self.maxpart))
            if not ok:
                fails += '* %d - %s\n' % (self.maxpart, output)
                #T:Partition %s konnte nicht entfernt werden
            self.maxpart -= 1
        res = self.freespace()
        if res:
            return res
        if self.maxpart >= partition0:
            return _("Partition deletion failed:\n") + fails
            #T:Entfernung der Partitionen schlug fehl
        return None


    def newpart(self, size, swap=False):
        """Create a new partition with the given size (MB) at the next possible
        location on the disk. With msdos partition tables no partition 4 will
        be used - this will be created as an extended partition covering the
        remainder of the drive. Further partitions will be "logical".
        """
        if swap:
            fstype = 'linux-swap'
        else:
            # This does not mean the partition will be formatted as ext2!
            fstype = 'ext2'
        if size > self.free_mb:
            return _("Not enough space on device")
            #T:Zu wenig Platz auf dem Gerät
        # Need extended? Is extended?
        ptype = 'primary'
        if (self.devinfo[3] ==  'msdos') and (self.maxpart >= 3):
            ptype = 'logical'
            if self.maxpart == 3:
                # Create an extended partition
                ok, output = run_capture('parted', self.device, 'unit', 'MB', 'mkpart',
                        'extended',  str(self.start_mb), str(mb2num(self.devinfo[1])))
                if not ok:
                    return _("Couldn't create extended partition:\n  ") + output
                    #T:Erweiterte Partition konnte nicht erstellt werden

        elif self.devinfo[3] != 'gpt':
            return _("Unsupported disk partition table: %s") % self.devinfo[3]
            #T:Partitionstabellentyp '%s' wird nicht unterstützt

        ok, output = run_capture('parted', self.device, 'unit', 'MB', 'mkpart',
                ptype, fstype,  str(self.start_mb), str(self.start_mb + size))
        if ok:
            return self.freespace()    # Reread disk info
        else:
            return _("Partition creation failed:\n  ") + output
        #T:Partition konnte nicht erstellt werden


    def format(self, label, dpath=None, swap=False):
        """Format the partition with the given label, if dpath is set use that.
        """
        dev = run_capture('blkid', '-L', label)[1].strip()
        if dpath:
            if dev and (dpath != dev):
                return _("Partition with label '%s' already exists") % label
        elif not dev:
            return _("Partition with label '%s' not found") % label

        fcmd = (['mkswap', '-L'] if swap
                else self.formatcommands.get(label, FORMATCOMMAND))
        fcmd.append(label)
        fcmd.append(dev)
        ok, output = run_capture(fcmd)
        return None if ok else output


    def get_partlist(self):
        return self.devinfo[3]


    def get_partinfo(self, i):
        istr = str(i)
        for p in self.devinfo[3]:
            if p[0]  == istr:
                return p
        return None


    def print_partitions(self):
        d = self.devinfo
        print d[0], d[1], d[2], d[4], d[5]+'/'+d[6]
        for p in d[3]:
            if (p[4] != p[5]) or (p[0] != u'-'):
                label = run_capture('blkid', '-s', 'LABEL', '-o', 'value', self.device+p[0])[1].strip()
                line = '%03s [%016s] %012s %016s' % (p[0], label, p[1], p[2])
                if p[3]:
                    line += '     (%s)' % p[3]
                print line
        print "+++++++++++++++++++++++++++"
        total = mb2num(self.devinfo[1])
        print _("Total: %d MB --- Free: %d MB") % (total, total - self.start_mb)
        #T:Insgesamt: %d GB --- Frei: %d GB


    def mount(self, src, dst, opts=""):
        if runcmd("mount %s %s %s" % (opts, src, dst))[0]:
            self._mounts.append(dst)
            return True
        return False


    def run_mount_devprocsys(self, script, *args):
        ok = True
        dirs = []
        for d in ('/dev', '/proc', '/sys'):
            mp = self.mountpoint + d
            if self.mount(d, mp, '--bind'):
                dirs.append(mp)
            else:
                ok = False
                break
        if ok:
            ok = scripts.run(script, *args)
        return self.unmount(dirs) and ok


    def unmount(self, dst=None):
        if dst == None:
            mnts = list(self._mounts)
        elif type(dst) in (list, tuple):
            mnts = list(dst)
        else:
            mnts = [dst]

        r = True
        for m in mnts:
            if runcmd("umount %s" % m)[0]:
                self._mounts.remove(m)
            else:
                r = False
        return r


    def sys_quit(self):
        self.unmount()


if __name__ == '__main__':
    class options:
        slave = False
        force = False
        quiet = False
    backend.init(options)
    part = Partitions('/dev/sda')
    part.print_partitions()
    sys_quit(0)
